# -*- coding: utf-8 -*-


from .event_handlers import EventStream, EventStreams, EventHandler  # noqa
from .event_types import events  # noqa
from .tasks import task  # noqa

event_handler = EventHandler.event_handler



# -*- coding: utf-8 -*-

import asyncio
import logging
import os
import sys

import peony

from . import exceptions

try:
    import magic
    mime = magic.Magic(mime=True)
except:  # pragma: no cover
    import mimetypes
    mime = mimetypes.MimeTypes()
    magic = None


_logger = logging.getLogger(__name__)


class Handle:

    __slots__ = 'exceptions', 'handler'

    def __init__(self, handler, *exceptions):
        self.exceptions = exceptions
        self.handler = handler


class MetaErrorHandler(type):

    def __new__(cls, name, bases, attrs, **kwargs):
        """ put the :class:`~peony.utils.Handle`s in the right place """
        handlers = {}

        for base in bases:
            if hasattr(base, '_handlers'):
                for key, value in base._handlers.items():
                    handlers[key] = value

        for attr in attrs.values():
            if isinstance(attr, Handle):
                for exception in attr.exceptions:
                    handlers[exception] = attr.handler

        attrs['_handlers'] = handlers

        return super().__new__(cls, name, bases, attrs)


class ErrorHandler(metaclass=MetaErrorHandler):
    """
        Basic error handler

    This error handler just raises all the exceptions that it receives.
    """

    RETRY = CONTINUE = OK = True
    RAISE = STOP = False

    def __init__(self, request):
        self.__request = request

    @staticmethod
    def handle(*exceptions):
        """  """
        def handler_decorator(handler):
            return Handle(handler, *exceptions)

        return handler_decorator

    async def __handle(self, exception_class, **kwargs):
        if exception_class in self._handlers:
            handler = self._handlers[exception_class]
            args = get_args(handler, skip=1)
            handler_kwargs = {key: kwargs[key] for key in args
                              if key in kwargs}
            try:
                return await execute(handler(self, **handler_kwargs))
            except Exception as exc:
                return exc

        for base in exception_class.__bases__:
            return await self.__handle(base, **kwargs)

        return ErrorHandler.RAISE

    async def __call__(self, future=None, **kwargs):
        while True:
            try:
                if future is None:
                    return await self.__request(**kwargs)
                else:
                    return await self.__request(future=future, **kwargs)
            except Exception as exc:
                reply = await self.__handle(exc.__class__, exception=exc,
                                            **kwargs)
                if isinstance(reply, Exception):
                    exc = reply
                if reply is not ErrorHandler.RETRY:
                    _logger.debug("raising exception")
                    if future is not None:
                        future.set_exception(exc)
                    raise exc


class DefaultErrorHandler(ErrorHandler):
    """
        The default error_handler

    The decorated request will retry infinitely on any handled error
    The exceptions handled are :class:`TimeoutError`,
    :class:`asyncio.TimeoutError`,
    :class:`exceptions.RateLimitExceeded` and
    :class:`exceptions.ServiceUnavailable`
    """

    def __init__(self, request):
        super().__init__(request)
        self.tries = 3

    @ErrorHandler.handle(exceptions.RateLimitExceeded)
    async def handle_rate_limits(self, exception, url):
        delay = int(exception.reset_in) + 1
        fmt = "Sleeping for {}s (rate limit exceeded on endpoint {})"
        _logger.warning(fmt.format(delay, url))
        await asyncio.sleep(delay)
        return ErrorHandler.RETRY

    @ErrorHandler.handle(asyncio.TimeoutError, TimeoutError)
    def handle_timeout_error(self, url):
        fmt = "Request to {url} timed out, retrying"
        _logger.info(fmt.format(url=url))
        return ErrorHandler.RETRY

    @ErrorHandler.handle(exceptions.ServiceUnavailable)
    async def handle_service_unavailable(self):
        self.tries -= 1
        if self.tries > 0:
            _logger.info("Service temporarily unavailable, "
                         "request will be made again very soon")
            await asyncio.sleep(0.5)
            return ErrorHandler.RETRY


def get_args(func, skip=0):
    """
        Hackish way to get the arguments of a function

    Parameters
    ----------
    func : callable
        Function to get the arguments from
    skip : int, optional
        Arguments to skip, defaults to 0 set it to 1 to skip the
        ``self`` argument of a method.

    Returns
    -------
    tuple
        Function's arguments
    """

    code = getattr(func, '__code__', None)
    if code is None:
        code = func.__call__.__code__

    return code.co_varnames[skip:code.co_argcount]


def log_error(msg=None, exc_info=None, logger=None, **kwargs):
    """
        log an exception and its traceback on the logger defined

    Parameters
    ----------
    msg : str, optional
        A message to add to the error
    exc_info : tuple
        Information about the current exception
    logger : logging.Logger
        logger to use
    """
    if logger is None:
        logger = _logger

    if not exc_info:
        exc_info = sys.exc_info()

    if msg is None:
        msg = ""

    exc_class, exc_msg, _ = exc_info

    if all(info is not None for info in exc_info):
        logger.error(msg, exc_info=exc_info)


async def get_media_metadata(data, path=None):
    """
        Get all the file's metadata and read any kind of file object

    Parameters
    ----------
    data : bytes
        first bytes of the file (the mimetype shoudl be guessed from the
        file headers
    path : str, optional
        path to the file

    Returns
    -------
    str
        The mimetype of the media
    str
        The category of the media on Twitter
    """
    if isinstance(data, bytes):
        media_type = await get_type(data, path)

    else:
        raise TypeError("get_metadata input must be a bytes")

    media_category = get_category(media_type)

    _logger.info("media_type: %s, media_category: %s" % (media_type,
                                                         media_category))

    return media_type, media_category


async def get_size(media):
    """
        Get the size of a file

    Parameters
    ----------
    media : file object
        The file object of the media

    Returns
    -------
    int
        The size of the file
    """
    if hasattr(media, 'seek'):
        await execute(media.seek(0, os.SEEK_END))
        size = await execute(media.tell())
        await execute(media.seek(0))
    elif hasattr(media, 'headers'):
        size = int(media.headers['Content-Length'])
    elif isinstance(media, bytes):
        size = len(media)
    else:
        raise TypeError("Can't get size of media of type:",
                        type(media).__name__)

    _logger.info("media size: %dB" % size)
    return size


async def get_type(media, path=None):
    """
    Parameters
    ----------
    media : file object
        A file object of the image
    path : str, optional
        The path to the file

    Returns
    -------
    str
        The mimetype of the media
    str
        The category of the media on Twitter
    """
    if magic:
        if not media:
            raise TypeError("Media data is empty")

        _logger.debug("guessing mimetype using magic")
        media_type = mime.from_buffer(media[:1024])
    else:
        media_type = None
        if path:
            _logger.debug("guessing mimetype using built-in module")
            media_type = mime.guess_type(path)[0]

        if media_type is None:
            msg = ("Could not guess the mimetype of the media.\n"
                   "Please consider installing python-magic\n"
                   "(pip3 install peony-twitter[magic])")
            raise RuntimeError(msg)

    return media_type


def get_category(media_type):
    if media_type.startswith('video'):
        return "tweet_video"
    elif media_type.endswith('gif'):
        return "tweet_gif"
    elif media_type.startswith('image'):
        return "tweet_image"
    else:
        raise RuntimeError("The provided media cannot be handled.\n"
                           "mimetype: %s" % media_type)


async def execute(coro):
    """
        run a function or coroutine

    Parameters
    ----------
    coro : asyncio.coroutine or function
    """
    if asyncio.iscoroutine(coro):
        return await coro
    else:
        return coro


def set_debug():
    """ activates error messages, useful during development """
    logging.basicConfig(level=logging.WARNING)
    peony.logger.setLevel(logging.DEBUG)



import json
import logging

from . import exceptions

logger = logging.getLogger(__name__)


class BaseJSONData(dict):
    """
        A dict in which you can access items as attributes

    >>> obj = JSONData(key=True)
    >>> obj['key'] is obj.key
    True
    """

    def __getattr__(self, key):
        if key in self:
            return self[key]

        raise AttributeError("%s has no property named %s." %
                             (self.__class__.__name__, key))

    def __delattr__(self, item):
        del self[item]

    def __setattr__(self, key, value):
        self[key] = value


class JSONData(BaseJSONData):
    """
        A dict that lets you get the full data of the tweet without having
        to check if the data is truncated
    """

    def __contains__(self, key):
        if key == 'text':
            return super().__contains__('text') or 'full_text' in self

        elif super().__contains__(key):
            return True

        if 'extended_tweet' in self.keys():
            return key in self.extended_tweet

        return False

    def __getitem__(self, key):
        if key == 'text' and 'full_text' in self.keys():
            return super().__getitem__('full_text')

        if key == 'extended_tweet':
            return super().__getitem__(key)

        if 'extended_tweet' in self.keys():
            if key in self.extended_tweet:
                return self.extended_tweet[key]

        return super().__getitem__(key)

    def get(self, key, default=None):
        # it seems like the get method still called another __getitem__
        # than that of the instance
        if key in self:
            return self[key]

        return default


class PeonyResponse:
    """
        Response objects

    In these object you can access the headers, the request, the url
    and the data
    getting an attribute/item of this object will get the corresponding
    attribute/item of the data

    >>> peonyresponse = PeonyResponse(
    ...     data=JSONData(key="test"), headers={},
    ...     url="http://google.com", request={}
    ... )
    >>> peonyresponse.key is peonyresponse.data.key  # returns True
    >>>
    >>> peonyresponse = PeonyResponse(
    ...     data=[JSONData(key="test"), JSONData(key=1)], headers={},
    ...     url="http://google.com", request={}
    ... )
    >>> # iterate over peonyresponse.data
    >>> for key in peonyresponse:
    ...     pass  # do whatever you want

    Parameters
    ----------
    data : JSONData, dict or list
        Data object
    headers : dict
        Headers of the response
    url : str
        URL of the request
    request : dict
        Requests arguments
    """

    def __init__(self, data, headers, url, request):
        super().__setattr__('data', data)
        super().__setattr__('headers', headers)
        super().__setattr__('url', url)
        super().__setattr__('request', request)

    def __getattr__(self, key):
        """ get attributes from the data """
        return getattr(self.data, key)

    def __getitem__(self, key):
        """ get items from the data """
        return self.data[key]

    def __contains__(self, item):
        return item in self.data

    def __iter__(self):
        """ iterate over the data """
        return iter(self.data)

    def __str__(self):
        """ use the string of the data """
        return str(self.data)

    def __repr__(self):
        """ use the representation of the data """
        return repr(self.data)

    def __len__(self):
        """ get the length of the data """
        return len(self.data)

    def __setitem__(self, key, value):
        self.data[key] = value
    __setattr__ = __setitem__

    def __delitem__(self, key):
        del self.data[key]
    __delattr__ = __delitem__


def loads(json_data, encoding="utf-8", **kwargs):
    """
        Custom loads function with an object_hook and automatic decoding

    Parameters
    ----------
    json_data : str
        The JSON data to decode
    *args
        Positional arguments, passed to :func:`json.loads`
    encoding : :obj:`str`, optional
        The encoding of the bytestring
    **kwargs
        Keyword arguments passed to :func:`json.loads`

    Returns
    -------
    :obj:`dict` or :obj:`list`
        Decoded json data
    """
    if isinstance(json_data, bytes):
        json_data = json_data.decode(encoding)

    return json.loads(json_data, object_hook=JSONData, **kwargs)


async def read(response, loads=loads, encoding=None):
    """
        read the data of the response

    Parameters
    ----------
    response : aiohttp.ClientResponse
        response
    loads : callable
        json loads function
    encoding : :obj:`str`, optional
        character encoding of the response, if set to None
        aiohttp should guess the right encoding

    Returns
    -------
    :obj:`bytes`, :obj:`str`, :obj:`dict` or :obj:`list`
        the data returned depends on the response
    """
    ctype = response.headers.get('Content-Type', "").lower()

    try:
        if "application/json" in ctype:
            logger.info("decoding data as json")
            return await response.json(encoding=encoding, loads=loads)

        if "text" in ctype:
            logger.info("decoding data as text")
            return await response.text(encoding=encoding)

    except (UnicodeDecodeError, json.JSONDecodeError) as exc:
        data = await response.read()
        raise exceptions.PeonyDecodeError(response=response,
                                          data=data,
                                          exception=exc)

    return await response.read()



# -*- coding: utf-8 -*-

import asyncio
import base64
import hmac
import random
import string
import time
import urllib.parse
from abc import ABC, abstractmethod
from hashlib import sha1

import aiohttp

from . import __version__, utils


def quote(s):
    return urllib.parse.quote(s, safe="")


class PeonyHeaders(ABC, dict):
    """
        Dynamic headers for Peony

    This is the base class of :class:`OAuth1Headers` and
    :class:`OAuth2Headers`.

    Parameters
    ----------
    compression : bool, optional
        If set to True the client will be able to receive compressed
        responses else it should not happen unless you provide the
        corresponding header when you make a request. Defaults to True.
    user_agent : str, optional
        The user agent set in the headers. Defaults to
        "peony v{version number}"
    headers : dict
        dict containing custom headers
    """

    def __init__(self, compression=True, user_agent=None, headers=None):
        """ Add a nice User-Agent """
        super().__init__()

        if user_agent is None:
            self['User-Agent'] = "peony v" + __version__
        else:
            self['User-Agent'] = user_agent

        if compression:
            self['Accept-Encoding'] = "deflate, gzip"

        if headers is not None:
            for key, value in headers.items():
                self[key] = value

    def __setitem__(self, key, value):
        super().__setitem__(key.title(), value)

    async def prepare_request(self, method, url,
                              headers=None,
                              skip_params=False,
                              proxy=None,
                              **kwargs):
        """
        prepare all the arguments for the request

        Parameters
        ----------
        method : str
            HTTP method used by the request
        url : str
            The url to request
        headers : dict, optional
            Additionnal headers
        proxy : str
            proxy of the request
        skip_params : bool
            Don't use the parameters to sign the request

        Returns
        -------
        dict
            Parameters of the request correctly formatted
        """

        if method.lower() == "post":
            key = 'data'
        else:
            key = 'params'

        if key in kwargs and not skip_params:
            request_params = {key: kwargs.pop(key)}
        else:
            request_params = {}

        request_params.update(dict(method=method.upper(), url=url))

        coro = self.sign(**request_params, skip_params=skip_params,
                         headers=headers)
        request_params['headers'] = await utils.execute(coro)
        request_params['proxy'] = proxy

        kwargs.update(request_params)

        return kwargs

    def _user_headers(self, headers=None):
        """ Make sure the user doesn't override the Authorization header """
        h = self.copy()

        if headers is not None:
            keys = set(headers.keys())
            if h.get('Authorization', False):
                keys -= {'Authorization'}

            for key in keys:
                h[key] = headers[key]

        return h

    @abstractmethod
    def sign(self, *args, headers=None, **kwargs):
        """
            sign, that is, generate the `Authorization` headers before making
            a request
        """


class OAuth1Headers(PeonyHeaders):
    """
        Dynamic headers implementing OAuth1

    Parameters
    ----------
    consumer_key : str
        Your consumer key
    consumer_secret : str
        Your consumer secret
    access_token : str
        Your access token
    access_token_secret : str
        Your access token secret
    **kwargs
        Other headers
    """

    def __init__(self, consumer_key, consumer_secret,
                 access_token=None, access_token_secret=None,
                 compression=True, user_agent=None, headers=None):
        super().__init__(compression, user_agent, headers)

        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.access_token = access_token
        self.access_token_secret = access_token_secret

        self.alphabet = string.ascii_letters + string.digits

    @staticmethod
    def _default_content_type(skip_params):
        if skip_params:
            return "application/octet-stream"
        else:
            return "application/x-www-form-urlencoded"

    def sign(self, method='GET', url=None,
             data=None,
             params=None,
             skip_params=False,
             headers=None,
             **kwargs):

        headers = self._user_headers(headers)

        if data:
            if 'Content-Type' not in headers:
                default = self._default_content_type(skip_params)
                headers['Content-Type'] = default

            params = data.copy()
        elif params:
            params = params.copy()

        oauth = {
            'oauth_consumer_key': self.consumer_key,
            'oauth_nonce': self.gen_nonce(),
            'oauth_signature_method': 'HMAC-SHA1',
            'oauth_timestamp': str(int(time.time())),
            'oauth_version': '1.0'
        }

        if self.access_token is not None:
            oauth['oauth_token'] = self.access_token

        oauth['oauth_signature'] = self.gen_signature(method=method, url=url,
                                                      params=params,
                                                      skip_params=skip_params,
                                                      oauth=oauth)

        headers['Authorization'] = "OAuth "

        for key, value in sorted(oauth.items(), key=lambda i: i[0]):
            if len(headers['Authorization']) > len("OAuth "):
                headers['Authorization'] += ", "

            headers['Authorization'] += quote(key) + '="' + quote(value) + '"'

        return headers

    def gen_nonce(self):
        return ''.join(random.choice(self.alphabet) for i in range(32))

    def gen_signature(self, method, url, params, skip_params, oauth):
        signature = method.upper() + '&' + quote(url) + '&'

        if params is None or skip_params:
            params = oauth
        else:
            params.update(oauth)

        param_string = ""

        for key, value in sorted(params.items(), key=lambda i: i[0]):
            if param_string:
                param_string += '&'

            param_string += quote(key) + '='

            if key == "q":
                encoded_value = urllib.parse.quote(value, safe="$:!?/()'*@")
                param_string += encoded_value
            else:
                param_string += quote(value)

        signature += quote(param_string)

        key = quote(self.consumer_secret).encode() + b"&"
        if self.access_token_secret is not None:
            key += quote(self.access_token_secret).encode()
        signature = hmac.new(key, signature.encode(), sha1)

        signature = base64.b64encode(signature.digest()).decode().rstrip("\n")
        return signature


class OAuth2Headers(PeonyHeaders):
    """
        Dynamic headers implementing OAuth2

    Parameters
    ----------
    consumer_key : str
        Your consumer key
    consumer_secret : str
        Your consumer secret
    client : .client.BasePeonyClient
        The client to authenticate
    bearer_token : :obj:`str`, optional
        Your bearer_token
    **kwargs
        Other headers
    """

    def __init__(self, consumer_key, consumer_secret, client,
                 bearer_token=None, compression=True, user_agent=None,
                 headers=None):
        super().__init__(compression, user_agent, headers)

        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.client = client
        self.basic_authorization = self.get_basic_authorization()
        self._refreshing = asyncio.Event()
        self._refreshing.clear()

        if bearer_token is not None:
            self.token = bearer_token

    async def sign(self, url=None, *args, headers=None, **kwargs):
        if url == self._invalidate_token.url():
            del self.token
        elif 'Authorization' not in self:
            await self.refresh_token()

        return self._user_headers(headers)

    def get_basic_authorization(self):
        creds = quote(self.consumer_key), quote(self.consumer_secret)
        keys = ':'.join(creds).encode('utf-8')

        auth = "Basic " + base64.b64encode(keys).decode('utf-8')

        return {'Authorization': auth,
                'Content-Type': "application/x-www-form-urlencoded;"
                                "charset=UTF-8"}

    @property
    def token(self):
        if 'Authorization' in self:
            return self['Authorization'][len("Bearer "):]

    @token.setter
    def token(self, access_token):
        self['Authorization'] = "Bearer " + access_token

    @token.deleter
    def token(self):
        del self['Authorization']

    @property
    def _invalidate_token(self):
        return self.client['api', '', ''].oauth2.invalidate_token

    async def invalidate_token(self):
        if 'Authorization' not in self:
            raise RuntimeError('There is no token to invalidate')

        token = self.token

        try:
            request = self._invalidate_token.post
            data = RawFormData({'access_token': token}, quote_fields=False)

            await request(_data=data, _headers=self.basic_authorization)
        except:
            self.token = token
            raise

    async def refresh_token(self):
        if self._refreshing.is_set():
            return await self._refreshing.wait()

        self._refreshing.set()

        if 'Authorization' in self:
            await self.invalidate_token()

        request = self.client['api', "", ""].oauth2.token.post
        token = await request(grant_type="client_credentials",
                              _headers=self.basic_authorization)

        self.token = token['access_token']

        self._refreshing.clear()


class RawFormData(aiohttp.FormData):

    def _gen_form_urlencoded(self):
        def key(item):
            return item[0]['name']

        data = ""
        for type_options, _, value in sorted(self._fields, key=key):
            if data:
                data += "&"

            data += "%s=%s" % (type_options['name'], value)

        charset = self._charset if self._charset is not None else 'utf-8'
        content_type = "application/x-www-form-urlencoded;charset=" + charset

        return aiohttp.payload.BytesPayload(data.encode(),
                                            content_type=content_type)



# -*- coding: utf-8 -*-
import abc

import peony.utils

from .commands import Commands
from .tasks import task


class EventHandler(task):

    def __init__(self, func, event, prefix=None, strict=False):
        super().__init__(func)

        self.prefix = prefix
        self.is_event = event

        if prefix is not None:
            self.command = Commands(prefix=prefix, strict=strict)

    def __call__(self, *args):
        argcount = self.__wrapped__.__code__.co_argcount

        if hasattr(self, 'command'):
            args = (*args, self.command,)

        args = args[:argcount]
        return super().__call__(*args)

    def __repr__(self):
        return "<{clsname}: event:{event} prefix:{prefix}>".format(
            clsname=self.__class__.__name__,
            prefix=self.prefix,
            event=self.is_event
        )

    @classmethod
    def event_handler(cls, event, prefix=None, **values):

        def decorator(func):
            event_handler = cls(
                func=func,
                event=event,
                prefix=prefix,
                **values
            )

            return event_handler

        return decorator


class EventStream(abc.ABC):

    def __init__(self, client):
        self._client = client
        self.functions = [getattr(self, func)
                          for func in dir(self) if self._check(func)]

        self.functions.sort(key=lambda i: getattr(i.is_event, 'priority', 0))

    def __getitem__(self, key):
        return self._client[key]

    def __getattr__(self, key):
        return getattr(self._client, key)

    @abc.abstractmethod
    def stream_request(self):
        pass

    async def start(self):
        if callable(self.stream_request):
            stream_request = self.stream_request()
        else:
            stream_request = self.stream_request

        while True:
            async with stream_request as resource:
                async for data in resource:
                    try:
                        await self._run(data)
                    except Exception:
                        msg = "error in %s._start:\n" % self.__class__.__name__
                        peony.utils.log_error(msg)

    def _check(self, func):
        if not func.startswith("_"):
            return isinstance(getattr(self, func), EventHandler)
        else:
            return False

    def _get(self, data):
        for event_handler in self.functions:
            argcount = len(peony.utils.get_args(event_handler.is_event))
            args = [data, self._client][:argcount]
            if event_handler.is_event(*args):
                return event_handler

    async def _run(self, data):
        event_handler = self._get(data)

        if event_handler:
            coro = event_handler(self, data)
            try:
                return await peony.utils.execute(coro)
            except:
                fmt = "error occurred while running {classname}.{handler}:"
                msg = fmt.format(classname=self.__class__.__name__,
                                 handler=event_handler.__name__)

                peony.utils.log_error(msg)


class EventStreams(list):

    def __init__(self):
        super().__init__()
        self.is_setup = False

    def check_setup(self, client):
        if not self.is_setup:
            self.setup(client)

    def get_tasks(self, client):
        self.check_setup(client)
        return [client.loop.create_task(stream.start()) for stream in self]

    def get_task(self, client):
        self.check_setup(client)
        if len(self) == 1:
            return client.loop.create_task(self[0].start())
        elif self:
            raise RuntimeError("more than 1 event stream")
        else:
            raise RuntimeError("no event stream")

    def setup(self, client):
        for i in range(len(self)):
            self[i] = self[i](client=client)

        self.is_setup = True



# -*- coding: utf-8 -*-


from functools import update_wrapper


class Task:

    def __init__(self, func):
        self.__wrapped__ = None
        update_wrapper(self, func)

    def __call__(self, *args, **kwargs):
        return self.__wrapped__(*args, **kwargs)

    def __str__(self):
        return "<{cls} {name}()>".format(cls=self.__class__.__name__,
                                         name=self.__wrapped__.__name__)

    def __repr__(self):
        return str(self)


task = Task



# -*- coding: utf-8 -*-
from time import time

from . import data_processing


def get_error(data):
    """ return the error if there is a corresponding exception """
    if isinstance(data, dict):
        if 'errors' in data:
            error = data['errors'][0]
        else:
            error = data.get('error', None)

        if isinstance(error, dict):
            if error.get('code') in errors:
                return error


async def throw(response, loads=None, encoding=None, **kwargs):
    """ Get the response data if possible and raise an exception """
    if loads is None:
        loads = data_processing.loads

    data = await data_processing.read(response, loads=loads,
                                      encoding=encoding)

    error = get_error(data)
    if error is not None:
        exception = errors[error['code']]
        raise exception(response=response, error=error, data=data, **kwargs)

    if response.status in statuses:
        exception = statuses[response.status]
        raise exception(response=response, data=data, **kwargs)

    # raise PeonyException if no specific exception was found
    raise PeonyException(response=response, data=data, **kwargs)


class PeonyException(Exception):
    """ Parent class of all the exceptions of Peony """

    def __init__(self, response=None, error=None, data=None, url=None,
                 message=None):
        """
            Add the response and data attributes

        Extract message from the error if not explicitly given
        """
        self.response = response
        self.data = data
        self.error = error
        self.url = url

        if not message:
            message = self.get_message()

        if url:
            message += "\nurl: " + url

        super().__init__(message)

    def get_message(self):
        if self.error is not None:
            return self.error.get('message', self.error)

        return str(self.data)


class PeonyDecodeError(PeonyException):

    def __init__(self, exception, *args, **kwargs):
        self.exception = exception
        super().__init__(*args, **kwargs)

    def get_message(self):
        return "Could not decode response data:\n%s" % self.data


class MediaProcessingError(PeonyException):
    pass


class StreamLimit(PeonyException):
    pass


class ErrorDict(dict):
    """ A dict to easily add exception associated to a code """

    def code(self, code):
        """ Decorator to associate a code to an exception """
        def decorator(exception):
            self[code] = exception
            return exception

        return decorator


statuses = ErrorDict()
errors = ErrorDict()


@errors.code(32)
class NotAuthenticated(PeonyException):
    pass


@errors.code(34)
class DoesNotExist(PeonyException):
    pass


@errors.code(64)
class AccountSuspended(PeonyException):
    pass


@errors.code(68)
class MigrateToNewAPI(PeonyException):
    pass


@errors.code(88)
class RateLimitExceeded(PeonyException):
    """ Exception raised on rate limit """

    @property
    def reset(self):
        """
            Time when the limit will be reset

        Returns
        -------
        int
            Time when the limit will be reset
        """
        return int(self.response.headers.get('X-Rate-Limit-Reset', 0))

    @property
    def reset_in(self):
        """
            Time in seconds until the limit will be reset

        Returns
        -------
        int
            Time in seconds until the limit will be reset
        """
        return max(self.reset - time(), 0)


@errors.code(92)
class SSLRequired(PeonyException):
    pass


@errors.code(130)
class OverCapacity(PeonyException):
    pass


@errors.code(131)
class InternalError(PeonyException):
    pass


@errors.code(135)
class CouldNotAuthenticate(PeonyException):
    pass


@errors.code(136)
class Blocked(PeonyException):
    pass


@errors.code(161)
class FollowLimit(PeonyException):
    pass


@errors.code(179)
class ProtectedTweet(PeonyException):
    pass


@errors.code(185)
class StatusLimit(PeonyException):
    pass


@errors.code(187)
class DuplicatedStatus(PeonyException):
    pass


@errors.code(215)
class BadAuthentication(PeonyException):
    pass


@errors.code(226)
class AutomatedRequest(PeonyException):
    pass


@errors.code(231)
class VerifyLogin(PeonyException):
    pass


@errors.code(251)
class RetiredEndpoint(PeonyException):
    pass


@errors.code(261)
class ReadOnlyApplication(PeonyException):
    pass


@errors.code(271)
class CannotMuteYourself(PeonyException):
    pass


@errors.code(272)
class NotMutingUser(PeonyException):
    pass


@errors.code(354)
class DMCharacterLimit(PeonyException):
    pass


@statuses.code(304)
class NotModified(PeonyException):
    pass


@statuses.code(400)
class BadRequest(PeonyException):
    pass


@statuses.code(401)
class Unauthorized(PeonyException):
    pass


@statuses.code(403)
class Forbidden(PeonyException):
    pass


@statuses.code(404)
class NotFound(PeonyException):
    pass


@statuses.code(406)
class NotAcceptable(PeonyException):
    pass


@statuses.code(410)
class Gone(PeonyException):
    pass


@statuses.code(415)
class UnsupportedMediaType(PeonyException):
    pass


@statuses.code(420)
class EnhanceYourCalm(PeonyException):
    pass


@statuses.code(422)
class UnprocessableEntity(PeonyException):
    pass


@statuses.code(429)
class TooManyRequests(PeonyException):
    pass


@statuses.code(500)
class InternalServerError(PeonyException):
    pass


@statuses.code(502)
class BadGateway(PeonyException):
    pass


@statuses.code(503)
class ServiceUnavailable(PeonyException):
    pass


@statuses.code(504)
class GatewayTimeout(PeonyException):
    pass
